/*
 (c) Copyright, Real-Time Innovations, $Date: 2006/09/16 23:58:20 $.
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.
*/

#ifndef PerfMon_h
#define PerfMon_h
#ifdef RTI_WIN32
  #include <pdh.h>
#endif
#include "osapi/osapi_type.h"

Peer class PerfMon {
  private:
    RTI_UINT32 _pid;
#ifdef RTI_LINUX
    int Cpu_tot;
    unsigned long long Hertz, prevtic;
    struct timeval oldtimev;

    struct {
        unsigned long long utime, stime; /* [jiffies] */
        unsigned long vsize; /* [bytes] */
    } mStat_process;

    struct NetDevStat {
        unsigned long byte_recv, packet_recv, err_recv, drop_recv, mcast_recv,
            byte_sent, packet_sent, err_sent, drop_sent, colls_sent;
    } mStatNetDev_current, mStatNetDev_initial;

    const char* ms_nic;
#elif defined(RTI_WIN32)
    HQUERY hQuery;
    HCOUNTER phCounter;
#endif

  public:
#ifdef RTI_WIN32
    PerfMon::PerfMon(const char* exe_name);
#else
    PerfMon(const char* nic, RTI_UINT32 pid = 0);
#endif
    
    ~PerfMon();
    int isValid();

    float getCpu();
    unsigned long getMemory();

    /* net dev statistic on the specified NIC */
    RTIBool getNetDevStat();

    unsigned long getByteReceived();
    unsigned long getPacketReceived();
    unsigned long getErrorReceived();
    unsigned long getDropReceived();
    unsigned long getMcastReceived();
    unsigned long getByteSent();
    unsigned long getPacketSent();
    unsigned long getErrorSent();
    unsigned long getDropSent();
    unsigned long getCollisions();

  private:
#ifdef RTI_LINUX
    RTIBool readProcStat();
    RTIBool readProcNetDev(struct NetDevStat*);
#endif
};

#endif/*PerfMon_h*/
