/*
 (c) Copyright, Real-Time Innovations, $Date: 2006/09/16 06:56:14 $.
 All rights reserved.
 No duplications, whole or partial, manual or electronic, may be made
 without express written permission.  Any such copies, or
 revisions thereof, must display this notice unaltered.
 This code contains trade secrets of Real-Time Innovations, Inc.
*/


class BatchedThroughputDataWriter {
private:
    ThroughputSequenceDataWriter *_dds_writer;
    ThroughputSequence _batch;
    int _currentElementsInBatch;
    int _maxElementsInBatch;
    Throughput *_througput_ptr_array[BATCH_SIZE_MAX];


public:
    BatchedThroughputDataWriter( ThroughputSequenceDataWriter *dds_writer, 
                                 int batch_size) 
    {
        _dds_writer = dds_writer;
        _maxElementsInBatch = batch_size;
        _currentElementsInBatch = 0;
    
	ThroughputSequence_initialize(&_batch);
        _batch.elem_seq.maximum(0);
	_batch.elem_seq.loan_discontiguous(_througput_ptr_array,
					   BATCH_SIZE_MAX, 
					   BATCH_SIZE_MAX);
    }


    DDS_InstanceHandle_t register_instance(Throughput &data) 
    {
        return DDS_HANDLE_NIL;
    }

    DDS_ReturnCode_t write(Throughput &data, DDS_InstanceHandle_t &handle) 
    {
        _througput_ptr_array[_currentElementsInBatch] = &data;
        ++_currentElementsInBatch;
        if ( _currentElementsInBatch % _maxElementsInBatch == 0 ) {
            return flush();
        }
	return DDS_RETCODE_OK;
    }

    DDS_ReturnCode_t flush()
    {
        _batch.elem_seq.length(_currentElementsInBatch);
        _currentElementsInBatch = 0;
        return _dds_writer->write(_batch, DDS_HANDLE_NIL);
    }
};


